@@ -8,6 +8,10 @@
 #include "pgrn-wal.h"
 #include "pgrn-writable.h"
 
+#ifdef PGRN_SUPPORT_TABLEAM
+#	include <access/tableam.h>
+#endif
+
 static bool PGrnWALEnabled = false;
 
 bool
@@ -2066,7 +2070,7 @@ pgroonga_wal_apply_all(PG_FUNCTION_ARGS)
 #ifdef PGRN_SUPPORT_WAL
 	LOCKMODE lock = AccessShareLock;
 	Relation indexes;
-	HeapScanDesc scan;
+	PGrnTableScanDesc scan;
 	HeapTuple indexTuple;
 
 	if (!PGrnIsWritable())
@@ -2078,8 +2082,8 @@ pgroonga_wal_apply_all(PG_FUNCTION_ARGS)
 						"while pgroonga.writable is false")));
 	}
 
-	indexes = heap_open(IndexRelationId, lock);
-	scan = heap_beginscan_catalog(indexes, 0, NULL);
+	indexes = pgrn_table_open(IndexRelationId, lock);
+	scan = pgrn_table_beginscan_catalog(indexes, 0, NULL);
 	while ((indexTuple = heap_getnext(scan, ForwardScanDirection)))
 	{
 		Form_pg_index indexForm = (Form_pg_index) GETSTRUCT(indexTuple);
@@ -2103,14 +2107,14 @@ pgroonga_wal_apply_all(PG_FUNCTION_ARGS)
 		{
 			RelationClose(index);
 			heap_endscan(scan);
-			heap_close(indexes, lock);
+			pgrn_table_close(indexes, lock);
 			PG_RE_THROW();
 		}
 		PG_END_TRY();
 		RelationClose(index);
 	}
 	heap_endscan(scan);
-	heap_close(indexes, lock);
+	pgrn_table_close(indexes, lock);
 #else
 	ereport(ERROR,
 			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
@@ -2126,6 +2130,7 @@ PGrnWALTruncate(Relation index)
 	int64_t nTruncatedBlocks = 0;
 	BlockNumber i;
 	BlockNumber nBlocks;
+	Buffer *buffers;
 	GenericXLogState *state;
 
 	LockRelation(index, PGrnWALLockMode());
@@ -2136,6 +2141,8 @@ PGrnWALTruncate(Relation index)
 		return 0;
 	}
 
+	buffers = palloc(sizeof(Buffer) * nBlocks);
+
 	state = GenericXLogStart(index);
 
 	{
@@ -2146,10 +2153,10 @@ PGrnWALTruncate(Relation index)
 		buffer = PGrnWALReadLockedBuffer(index,
 										 PGRN_WAL_META_PAGE_BLOCK_NUMBER,
 										 BUFFER_LOCK_EXCLUSIVE);
+		buffers[0] = buffer;
 		page = GenericXLogRegisterBuffer(state, buffer, GENERIC_XLOG_FULL_IMAGE);
 		pageSpecial = (PGrnWALMetaPageSpecial *) PageGetSpecialPointer(page);
 		pageSpecial->next = PGRN_WAL_META_PAGE_BLOCK_NUMBER + 1;
-		UnlockReleaseBuffer(buffer);
 
 		nTruncatedBlocks++;
 	}
@@ -2160,6 +2167,7 @@ PGrnWALTruncate(Relation index)
 		Page page;
 
 		buffer = PGrnWALReadLockedBuffer(index, i, BUFFER_LOCK_EXCLUSIVE);
+		buffers[i] = buffer;
 		if (nTruncatedBlocks >= MAX_GENERIC_XLOG_PAGES)
 		{
 			GenericXLogFinish(state);
@@ -2167,13 +2175,18 @@ PGrnWALTruncate(Relation index)
 		}
 		page = GenericXLogRegisterBuffer(state, buffer, GENERIC_XLOG_FULL_IMAGE);
 		PageInit(page, BLCKSZ, 0);
-		UnlockReleaseBuffer(buffer);
 
 		nTruncatedBlocks++;
 	}
 
 	GenericXLogFinish(state);
 
+	for (i = 0; i < nBlocks; i++)
+	{
+		UnlockReleaseBuffer(buffers[i]);
+	}
+	pfree(buffers);
+
 	PGrnIndexStatusSetWALAppliedPosition(index,
 										 PGRN_WAL_META_PAGE_BLOCK_NUMBER + 1,
 										 0);
@@ -2244,11 +2257,11 @@ pgroonga_wal_truncate_all(PG_FUNCTION_ARGS)
 #ifdef PGRN_SUPPORT_WAL
 	LOCKMODE lock = AccessShareLock;
 	Relation indexes;
-	HeapScanDesc scan;
+	PGrnTableScanDesc scan;
 	HeapTuple indexTuple;
 
-	indexes = heap_open(IndexRelationId, lock);
-	scan = heap_beginscan_catalog(indexes, 0, NULL);
+	indexes = pgrn_table_open(IndexRelationId, lock);
+	scan = pgrn_table_beginscan_catalog(indexes, 0, NULL);
 	while ((indexTuple = heap_getnext(scan, ForwardScanDirection)))
 	{
 		Form_pg_index indexForm = (Form_pg_index) GETSTRUCT(indexTuple);
@@ -2272,14 +2285,14 @@ pgroonga_wal_truncate_all(PG_FUNCTION_ARGS)
 		{
 			RelationClose(index);
 			heap_endscan(scan);
-			heap_close(indexes, lock);
+			pgrn_table_close(indexes, lock);
 			PG_RE_THROW();
 		}
 		PG_END_TRY();
 		RelationClose(index);
 	}
 	heap_endscan(scan);
-	heap_close(indexes, lock);
+	pgrn_table_close(indexes, lock);
 #else
 	ereport(ERROR,
 			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
